#!/usr/bin/python
import subprocess
import string
import os
import os.path
import sys
import ConfigParser

abs_path = os.path.abspath(sys.argv[0])
basename = os.path.basename(sys.argv[0])
dirname = abs_path.replace(basename, "")
if "diskmon/bin" in dirname:
   if not "%slib" % dirname[:-4] in sys.path:
      sys.path.insert(0,"%slib" % dirname[:-4]) # in case you run this in benchmark directory as diskmon/bin/ana_snmptrap_sender.py --anaDisks e.g.
if not "%s/diskmon/lib" % dirname in sys.path:
    sys.path.insert(0,"%sdiskmon/lib" % dirname) # always with regard to parent script
if not "/usr/bp/lib/monitoring" in sys.path:
    sys.path.append("/usr/bp/lib/monitoring")
    
from bpdb     import *   # include config << bpdb as well
from optparse import OptionParser
parser = OptionParser()
parser.add_option("-d", "--debug",   dest="debug", default=False, action='store_true',
                  help="debug option to show detail commands to test")
parser.add_option("-v", "--verbose", dest="verbose", default=False, action='store_true',
                  help="verbose option to show output for debug purpose")
parser.add_option("-f", "--force", dest="force", default=False, action='store_true',
                  help="sending trap ignoring pass threshold (Sunday midnight)")
parser.add_option("-a", "--ana",   dest="ana", default=False, action='store_true',
                  help="run all the ana functions")
parser.add_option("--anaProtected",   dest="anaProtected", default=False, action='store_true',
                  help="analysis to protected number of agents/applications and total amount of files over file level/VMESX/HyperV/Exchage/MSSQL/Sharepoint/Oracle")
parser.add_option("--anaEsxServers",  dest="anaEsxServers", default=False, action='store_true',
                  help="analysis to get ESX server version information")
parser.add_option("--anaCloudHook",   dest="anaCloudHook", default=False, action='store_true',
                  help="analysis to get cloudhook information")
parser.add_option("--anaDisks",   dest="anaDisks", default=False, action='store_true',
                  help="analysis to get disk health information")
parser.add_option("--anaDatabase",   dest="anaDatabase", default=False, action='store_true',
                  help="analysis to get database health")
parser.add_option("-c", "--confirm",   dest="confirm", default=False, action='store_true',
                  help="Without confirm flag, snmp won't be sent")
(opts, args) = parser.parse_args()

CONFIG.debug        =opts.debug
CONFIG.verbose      =opts.verbose
CONFIG.confirm      =opts.confirm
CONFIG.anaProtected =opts.anaProtected
CONFIG.anaEsxServers=opts.anaEsxServers
CONFIG.anaCloudHook =opts.anaCloudHook
CONFIG.anaDisks     =opts.anaDisks
CONFIG.anaDatabase  =opts.anaDatabase 
CONFIG.force        =opts.force

if opts.ana:
    CONFIG.anaProtected =True
    CONFIG.anaEsxServers=True
    CONFIG.anaCloudHook =True
    CONFIG.anaDisks     =True
    CONFIG.anaDatabase  =True

CONFIG.nAsset =0  # Number of assets in node table:
                  # if they are, then they should have "System Metadata" as an application_instance.
                  # Also, associated systems (replication source) can have it 
    
from trap   import *   # include config << bpdb as well
from config import get_master_ini,get_first_word,get_printout
from util   import get_queryResult, binary_bit_lifter, get_hex, does_exist

CONFIG.db_accessible =True
CONFIG.dbOff_reason=""
def is_db_accessible():
   reason=""
   import commands      
   lines = commands.getoutput("/bin/ps aux | /bin/grep postgres | /bin/grep data").splitlines()
   dbON     =True
   # psql: could not fork new process for connection: Cannot allocate memory
   # True 1 ['root      9314  0.0  0.0  63848  1048 pts/2    S+   09:34   0:00 sh -c { ps aux | grep postgres | grep data; } 2>&1']

   lcnt = 0
   for line in lines:
      if "grep" in line:       continue
      lcnt +=1
   if not lcnt:
      dbON = False
      reason="no process postgres running"
      return reason
   
   # without bpdb postgres, you will see the following message from crontab:
   # psql: FATAL:  role "root" does not exist
   lines= commands.getoutput("/usr/bin/psql bpdb postgres -c \"select * from bp.nodes limit 1\"").splitlines()
   for line in lines:
      if "could not access status of transaction" in line or \
         "timeout expired" in line or \
         "Connection refused" in line or \
         "the database system is starting up" in line or \
         "too many clients" in line:
         dbON = False
         reason=line[:-1] # getting rid of return character (\n)
         return reason
      
   # Do not run while VACUUM is running
   lines = commands.getoutput("/bin/ps aux | /bin/grep VACUUM | /bin/grep -v grep").splitlines()
   if len(lines):
      dbON=False
      line=lines[0]
      reason=line[:-1]
      return reason
   
   return reason
      
def get_public_ip():
   #162.216.32.66 #DSM
   #64.64.199.*   #Chicago
   command= "/usr/bin/wget http://icanhazip.com -o /tmp/.wget.log -O /tmp/.public_ip"
   os.system(command)
   public_ip=get_first_word("cat /tmp/.public_ip")
   os.system("rm -rf /tmp/.public_ip")

   return public_ip


def get_hash_numbers():        
    nRef,nHashes=0,0
    if "ERROR" == get_master_ini("fileDedup","MetadataDirectory"):
        query   ="SELECT count(1) FROM bp.hash_values"
        if opts.debug: print query
        output =get_queryResult(query)[0]
        #psql: timeout expired
        if "timeout" in output: nRef,nHashes=-1,-1
        else:
            nHashes =int(output)
            query   ="SELECT count(1) FROM bp.backups_hash_values"
            if opts.debug: print query
            nRef    =int(get_queryResult(query)[0])
    else:
        nRef=int(get_first_word("/usr/bp/bin/metadataUtil -b | while read b; do metadataUtil -c$b; done | awk '{s+=$1} END {print s}'"))
        nHashes=int(get_first_word("/usr/bp/bin/metadataUtil -L | wc -l"))
        
    return (nHashes,nRef)

def get_d2d_utilization():
    physical_consumption=0 # du -ks shows number of KB
    for d2d in get_queryResult("SELECT DISTINCT dev_nr_name FROM bp.devices WHERE device_type=128"):
        du_ks=int(get_first_word("/usr/bin/du -ks %s" % d2d))
        physical_consumption +=du_ks
        
    sum_total_megs=0
    sum_logical_size_backups=0
    sum_hash_compressed_length=0
    sum_hash_length=0
    return


def main():
    CONFIG.dbOff_reason=is_db_accessible()   
    CONFIG.db_accessible = not len(CONFIG.dbOff_reason)
   
    ####
    #   See the README.snmptrap for documentation
    ######
    message = ''
    category_id = 18 # New informational trap
    severity=3 #informational
    isopen = 0
    doSendInfoTrap=True
    if CONFIG.anaEsxServers:
        message =getAnaEsxServers()
        if len(message):    
            shortName="AnaESXservers"
            trapToSend = Trap(category_id, shortName, message, severity, isopen, 206) # This is for version information trap
            print trapToSend.message
            if does_pass_threshold(shortName,6) or CONFIG.force:  #Crontab configured to send every 7 days; 6 days is set just to prevent edge case
                trapToSend.send_trap(doSendInfoTrap)

    if CONFIG.anaProtected:
        message =getAnaProtected()
        if len(message):    
            shortName="AnaProtected"
            trapToSend = Trap(category_id, shortName, message, severity, isopen, 206)
            print trapToSend.message
            if does_pass_threshold(shortName,6) or CONFIG.force:  #Crontab configured to send every 7 days; 6 days is set just to prevent edge case
                trapToSend.send_trap(doSendInfoTrap) #Default is not sending a trap if it's informational trap


    if CONFIG.anaCloudHook:
        message =getAnaCloudHook()
        if len(message):    
            shortName="AnaCloudHook"
            trapToSend = Trap(category_id, shortName, message, severity, isopen, 206) # This is for version information trap
            print trapToSend.message
            if does_pass_threshold(shortName,6) or CONFIG.force:  #Crontab configured to send every 7 days; 6 days is set just to prevent edge case
                trapToSend.send_trap(doSendInfoTrap)

    if CONFIG.anaDisks:
        diskmon_file="/var/tmp/disk_mon.out.bak"
        if not os.path.exists(diskmon_file):
            diskmon_file="/var/tmp/disk_mon.out"

        messages =getAnaDisks(diskmon_file) # number of intakable characters in notification server is 1001
                                # each drive consumes 40 chars (Seagate), so I decided to split every 20 drives 
        for i in range(len(messages)):
            message=messages[i]    
            if len(message):    
                shortName="AnaDisks"
                trapToSend = Trap(category_id, shortName, message, severity, isopen, 206) # This is for version information trap
                print trapToSend.message
                if does_pass_threshold(shortName,6) or CONFIG.force:  #Crontab configured to send every 7 days; 6 days is set just to prevent edge case
                    trapToSend.send_trap(doSendInfoTrap)
                
    if CONFIG.anaDatabase:
        message =getAnaDatabase()
        if len(message):    
            shortName="AnaDatabase"
            trapToSend = Trap(category_id, shortName, message, severity, isopen, 206) # This is for version information trap
            print trapToSend.message
            if does_pass_threshold(shortName,6) or CONFIG.force:  #Crontab configured to send every 7 days; 6 days is set just to prevent edge case
                trapToSend.send_trap(doSendInfoTrap)
    return

def getAnaDatabase():
   
   message ="[INFO] Ana Database "

   ui_flag="rrc"
   lines=get_printout("/bin/grep ui: /var/opt/unitrends/platform_capabilities | /bin/grep -v \#")
   if len(lines):
      if "satori" in lines[0]:
         ui_flag="satori"
   message+="%s " %ui_flag       
   
   ret=get_printout("cd /usr/bp/data/; /bin/df -P . | /usr/bin/tail -1")
   part_total_size_gib=0
   part_used_size_gib=0
   if len(ret):
      words=ret[0].split()
      part_total_size_gib=int( int(words[1])/1000./1000. + 0.5 )
      part_used_size_gib =int( int(words[2])/1000./1000. + 0.5 )

   postgres_used_size_gib  =int(get_first_word("/usr/bin/du -ks /usr/bp/data/"))
   postgres_used_size_gib  =int(postgres_used_size_gib/1000./1000.+0.5)

   part_total_size_gib_hex    =get_hex(part_total_size_gib,5)  #up to 1,048,576 GiB
   part_used_size_gib_hex     =get_hex(part_used_size_gib,5)
   postgres_used_size_gib_hex =get_hex(postgres_used_size_gib,4)
   
   # capture db related
   message+="%s %s %s " % (part_total_size_gib_hex,part_used_size_gib_hex,postgres_used_size_gib_hex)

   du_ks_used_metadata_gib=0   
   list_major_tables=['hash_values','backups_hash_values','backups','dirs','backup_files']
   list_found_major_tables=[]
   str_major_tables=''
   if get_master_ini("fileDedup","MetadataGeneration")=="1":
      metadata_path=get_master_ini("fileDedup","MetadataDirectory")
      list_major_tables=['backups','dirs','backup_files']
      if metadata_path != "ERROR":
         du_ks_used_metadata_gib  =int(get_first_word("/usr/bin/du -ks %s" % metadata_path))
   du_ks_used_metadata_gib_hex=get_hex(du_ks_used_metadata_gib,4) # up to 65536: 65 TiB   
   message +="|N %s" %du_ks_used_metadata_gib_hex
      
   if not CONFIG.db_accessible:
      message += "| DB is not accessible because %s" % CONFIG.dbOff_reason 
      return message
   
   for i in range(len(list_major_tables)):
      major_table=list_major_tables[i]
      str_major_tables+="'%s',"%major_table
   str_major_tables=str_major_tables[:-1]   
   
   psql_command ="/usr/bin/psql bpdb postgres -tq -c \""
   psql_command+="SELECT relname FROM bp.bloat_tables_bpdb "
   psql_command+="WHERE  relname in (%s)" % str_major_tables
   psql_command+="\" | grep -v ^$"
   if opts.debug: print psql_command
   for line in get_printout(psql_command):
      list_found_major_tables.append(line.strip())  # strip take care of the first white space in the line

   # it translates into the difference of the first from the second.
   # eg.
   #>>> b=['backups', 'backup_files', 'dirs']
   #>>> a=['hash_values', 'backups_hash_values', 'backups', 'dirs', 'backup_files']
   #>>> set(b).difference(a)
   #set([])
   #>>> set(a).difference(b)
   #set(['hash_values', 'backups_hash_values'])
   
   missing_tables=set(list_major_tables).difference(list_found_major_tables) 
   missing_tables=list(missing_tables)

   #missing_tables=['dirs'] # for test

   for table in missing_tables:
      psql_command ="/usr/bin/psql bpdb postgres -tq -c \""
      psql_command+="analyze verbose bp.%s" % table
      psql_command+="\""
      print psql_command
      os.system(psql_command)

   limit_number_of_tables=8   
   query="""
   SELECT bdw.relname,
   cc.reltuples::bigint/1000 as rows,
   (cc.relpages::bigint)*8/1024 as current_size_mb,
   (ceil(cc.reltuples*bdw.datawidth/current_setting('block_size')::NUMERIC)::bigint)*8/1024 AS expected_size_mb,
   db_info.db_size
   FROM (
   SELECT
         pg_size_pretty(pg_database_size(pg_database.datname)) AS db_size
         FROM pg_database where datname= 'bpdb'
         ) db_info
      ,bp.bloat_datawidth_bpdb bdw JOIN pg_class cc ON cc.oid = bdw.relid
      AND bdw.schemaname = 'bp' AND (bdw.datawidth IS NOT NULL or bdw.datawidth > 0)
   ORDER BY expected_size_mb DESC LIMIT %d;
   """  % limit_number_of_tables
   psql_command  ="/usr/bin/psql bpdb postgres -tq -c \"%s\" | grep -v ^$" % query

   db_size_gib=0
   tuples_db=[]
   lines =get_printout(psql_command)

   for i in range(len(lines)):
      line=lines[i]
      words=line.split("|")
      if opts.debug:  print words
      if i==0:
         segs=words[4].split()
         db_size_gib=int(segs[0]) # mostly GiB
         if segs[1].strip() == "TB":
            db_size_gib=int(segs[0])*1024
         elif segs[1].strip() == "MB":
            db_size_gib=int(0.5+int(segs[0])/1024.)
         
      table_name=words[0].strip()
      n_kilo_rows=int(int(words[1])/1000.+0.5)
      cur_mib=int(words[2])
      exp_mib=int(words[3])
      
      tuples_db.append( (table_name,n_kilo_rows,cur_mib,exp_mib) )

   db_size_gib_hex            =get_hex(db_size_gib,4) # up to 65536: 65TiB
   message+="|D %s" % db_size_gib_hex

   if opts.debug:
      print "partition total/used/postgres used/bpdb used in GB = %d, %d, %d, %d" %(part_total_size_gib,part_used_size_gib,postgres_used_size_gib,db_size_gib)

   
   psql_command ="/usr/bin/psql bpdb postgres -tq -c \""
   psql_command+="SELECT dev_rw_name from bp.devices "
   psql_command+="WHERE comment NOT LIKE '%%Tape%%'"
   psql_command+="\" | grep -v ^$"
   if opts.debug: print psql_command
   tuples_d2d=[]
   for line in get_printout(psql_command):
      D2D_path=line.strip()
      ret=get_printout("cd %s; /bin/df -P . | /usr/bin/tail -1" % D2D_path)
      total_size_gib=0
      used_size_gib=0
      if len(ret):
         words=ret[0].split()
         total_size_gib=int( int(words[1])/1000./1000. + 0.5 )
         used_size_gib =int( int(words[2])/1000./1000. + 0.5 )
      du_ks_used_size_gib  =int(get_first_word("/usr/bin/du -ks %s" % D2D_path))
      du_ks_used_size_gib  =int(du_ks_used_size_gib/1000./1000.+0.5)
      
      tuples_d2d.append( (D2D_path, total_size_gib, used_size_gib, du_ks_used_size_gib) )

   for tup in tuples_d2d:   
      if opts.debug: print "tuples_d2d: ", tup
      D2D_path=tup[0]
      total_size_gib_hex=get_hex(tup[1],6)
      used_size_gib_hex =get_hex(tup[2],6)
      du_ks_size_gib_hex=get_hex(tup[3],6)
      message +="|* %s %s %s %s " %(D2D_path,total_size_gib_hex,used_size_gib_hex,du_ks_size_gib_hex)

   for tup in tuples_db:
      if opts.debug: print "tuples_db: ", tup
      table_name=tup[0]
      n_kilo_rows_hex=get_hex(tup[1],6)  # up to 16,777,216
      cur_mib_hex    =get_hex(tup[2],6)
      exp_mib_hex    =get_hex(tup[3],6)
      message+="|t %s %s %s %s" % (table_name,n_kilo_rows_hex,cur_mib_hex,exp_mib_hex) 
   
   return message    
    

 
def getAnaDisks(diskmon_file):

    header ="[INFO] Disks Snapshot" 

    if does_exist("grep 'This is UEB' %s" %diskmon_file):
        return getAnaDisks_UEB(diskmon_file,header)
        
    is_LSI_controller=does_exist("grep LSI %s" % diskmon_file)
    is_3ware_controller=does_exist("grep 'HW Controller:' %s | grep -E '9750-|9650SE-|9550'" % diskmon_file)
    asset_tag=get_first_word("cat /etc/unitrends-asset")
    appliance_model=asset_tag.split("-")[0]    
    drives =get_printout("cat %s | grep GB" % diskmon_file)
  
    if is_LSI_controller:
        return getAnaDisks_LSI(drives, header)
    elif is_3ware_controller:
        return getAnaDisks_3ware(drives, header)
    else:
        return getAnaDisks_Mdadm(drives, header)
        
def getAnaDisks_UEB(diskmon_file,header):

    # CPU Information: n(CPU)=1 aesON=False / Intel(R) Xeon(R) CPU E5620 @ 2.40GHz
    # RAM Information: 4 GB
    # ** /dev/sda7                 (     39 GiB )
    # ** /dev/sdb                  (    128 GiB )
    # ** /dev/sdc                  (   1024 GiB )
               
    messages=[]
    message ="%s UEB " % header

    for line in get_printout("/usr/bp/bin/bptool -v 2>&1| grep -E 'Feature:|Exp Date:|Inst Date:'"):
       words=line.split(":  ")
       if "Date:" in line:
          message+="|%s" % (words[1].strip()) #[4:]
       else:
          string=words[1].strip()
          string=string.replace(",",".") # to avoid collision with CSV
          message+="|%s" % (string) #[4:]
    message+="--"          
    
    hyperVisor,nCPU,is_aes,cpu_model,ram_GB="",0,0,"",0
    device_capacities=[]
    for line in get_printout("grep -E 'This is UEB|CPU|RAM|\*\*' %s" %diskmon_file):
        words=line.split()
        if "This is UEB" in line:
            hyperVisor=words[5].strip()
        if "CPU" in line:
            nCPU =int((words[2]).split("=")[1])
            if "True" in words[3]:
                is_aes=1
            cpu_model =line.split("/")[1].strip()
        if "RAM" in line:
            ram_GB=int(words[2])
        if "**" in line:
            device_capacities.append(int(words[3]))
            
    nCPU_hex=get_hex(nCPU,3)  # up to 4096 threads
    ram_GB_hex=get_hex(ram_GB,4) # up to 65536GB
                     
    message+="%s|%s%s%d|%s|" %(hyperVisor,nCPU_hex,ram_GB_hex,is_aes,cpu_model)

    for dev in device_capacities:
        message+="%s"%get_hex(dev,5) # up to 1048576 GB
    
    messages.append(message)
    return messages

def getAnaDisks_3ware(drives, header):
     
    messages=[]
    message="%s 1/1 " % header
    for i in range(len(drives)):
        line=drives[i]
        words=line.split()
        bay_no=0
        if "bay" in (words[2]).lower():
            bay_no=int((words[2]).split("#")[1])
        if "OS" in words[2]:
            bay_no=253
        if "EX" in words[2]:
            bay_no=254
        if "SSD" in words[15]:
            bay_no=255
        message+=get_string(words,bay_no)
        
    messages.append(message[:-1]) # getting rid of the last comma
    return messages   

def getAnaDisks_Mdadm(drives, header):
     
    messages=[]
    message="%s 1/1 " % header
    for i in range(len(drives)):
        line=drives[i]
        words=line.split()
        bay_no=0
        if "bay" in (words[2]).lower():
            bay_no=int((words[2]).split("#")[1])
        if "SSD" in words[15]:
            bay_no=255 
        message+=get_string(words,bay_no)
        
    messages.append(message[:-1]) # getting rid of the last comma
    return messages   

def get_string(words,bay_no):
    POD=int(words[6]) # power on days
    PND=int(words[8])
    if PND>255: PND=255
    UNC=int(words[9])
    if UNC>255: UNC=255
    RLC=int(words[10])
    
    SKE=int( (words[11]).split("/")[0] )
    WTB=int( (words[13])[:-1] )
    if WTB<0: WTB=0
    RTB=int(words[14])
    if RTB<0: RTB=0
    SN,FW=words[16],words[17]
    MD,ref_Idx='',0  # MD stands for Model Index
    if "0x" in words[21]:
       MD,ref_Idx=(words[21])[2:],21
    elif "0x" in words[20]:
       MD,ref_Idx=(words[20])[2:],20
    RER=int( words[ref_Idx+4] )  # read error rate count: max=3
    if RER>3: RER=3
    TMP=int( words[ref_Idx+1] )
    if TMP>63: TMP=63
    RER_TMP= binary_bit_lifter(RER,7)+ TMP
    CTO=int( words[ref_Idx+10] ) #command timeout
    import math
    if CTO>0:
       CTO= math.log(float(CTO),10)
       CTO=int(CTO+0.5)
    if CTO>15: CTO=15
    
    POD_hex=get_hex(POD,4)
    PND_hex=get_hex(PND,2)
    UNC_hex=get_hex(UNC,2)
    RLC_hex=get_hex(RLC,4)
    SKE_hex=get_hex(SKE,3)
    WTB_hex=get_hex(WTB,3)
    RTB_hex=get_hex(RTB,3)
    RER_TMP_hex=get_hex(RER_TMP,2)
    bay_no_hex=get_hex(bay_no,2)
    CTO_hex=get_hex(CTO,1)
    return "%s:%s:%s%s%s%s%s%s%s%s%s%s%s,"%(SN,FW,POD_hex,RLC_hex,PND_hex,UNC_hex,SKE_hex,WTB_hex,RTB_hex,RER_TMP_hex,MD,bay_no_hex,CTO_hex)
    #return "%s:%s:%s%s%s%s%s%s%s%s%s%s,"%(SN,FW,POD_hex,RLC_hex,PND_hex,UNC_hex,SKE_hex,WTB_hex,RTB_hex,RER_TMP_hex,MD,bay_no_hex)


def getAnaDisks_LSI(drives, header):
     
    messages=[]

    message="%s 1/1 " % header
    pre_enc_dev_id,increment=-1,0
    if len(drives)>20:
        message="%s 1/2 " % header

    for i in range(len(drives)):
        if i==20:
            messages.append(message[:-1]) # getting rid of the last comma
            message="%s 2/2 " % header
        
        line=drives[i]
        words=line.split()
        braket=words[20]
        enc_dev_id,bay_no=-1,-1
        if "[" in braket:
            segs=braket[1:-1].split(":")
            if len(segs[0]):
                enc_dev_id=int(segs[0])
            slot_no=int(segs[1])
      
            if enc_dev_id!=pre_enc_dev_id: 
                if pre_enc_dev_id>0: increment=24 
                pre_enc_dev_id=enc_dev_id
            if enc_dev_id>=0:
                bay_no=increment+slot_no
            if "SSD" in words[15]:
                bay_no=255
        else:
            if "OS" in words[2]:
                bay_no=253
            
        message+=get_string(words,bay_no)    
    messages.append(message[:-1]) # getting rid of the last comma
    return messages   

    

def getAnaCloudHook():
    enum = {'cloud:s3':0, 'cloud:swift':1, 'cloud:gs':2}
    cnts = []
    cnts_current = []
    max_gbs =[]
    used_gbs = []
    n_found =0
    for i in range(len(enum)):  # initialize
        cnts.append(0)
        cnts_current.append(0)
        max_gbs.append(0)
        used_gbs.append(0)
        
    for line in get_printout("/usr/bp/bin/cloudadm status 2>&1| /bin/grep cloud:"):
        # gathering data
        words = line.split("|")
        vendor = words[0].strip()
        online = True
        if words[2].strip() == "f":
            online = False
        max_gb  = int(words[3].strip())    
        used_gb = int(words[4].strip())    

        # measuring
        n_found +=1
        cnts[enum[vendor]] +=1
        if online:
            cnts_current[enum[vendor]] +=1
        max_gbs[enum[vendor]]  += max_gb
        used_gbs[enum[vendor]] += used_gb

    if not n_found:
        return ''

    message = "[INFO] Cloudhook information "
    for i in range(len(enum)):
        theKey =''
        for key,value in enum.items():
            if i==value: theKey = key
        message += "[%s](%d,%d,%dG,%dG) " %( theKey, cnts[i], cnts_current[i], max_gbs[i], used_gbs[i])

    return message    
    

def getAnaEsxServers():
    message ="[INFO] Esx Servers "
    if not CONFIG.db_accessible:
       message += " DB is not accessible because %s" % CONFIG.dbOff_reason 
       return message

    ####
    vmware_esx_servers = get_db_entries("vmware_esx_servers")
    enum ={'3.5.0':0, '4.0.0':1, '4.1.0':2, '5.0.0':3, '5.1.0':4, '5.5.0':5, '6.0.0':6}
    enum_hv ={'VMware':0, 'HyperV':1, 'XenServer':2, 'Else':3} 
    cnts, cnts_current, total_num_cpu_sockets =[],[],[]
    n_current, n_found=0,0
    
    for i in range(len(enum)):
        cnts.append(0)
        cnts_current.append(0)

    for i in range(len(enum_hv)):
        total_num_cpu_sockets.append(0)

    for server in vmware_esx_servers:
        cnts[enum[server.version]] += 1
        n_found +=1
        if server.current:
            total_num_cpu_sockets[enum_hv['VMware']]+=server.num_cpu_sockets
            n_current +=1
            cnts_current[enum[server.version]] += 1

    ####  XenServer part: version as well as num_cpu_sockets
    enum_xenver={'6.2.0':0,'6.5.0':1}
    cnt_xen=[]
    for i in range(len(enum_xenver)*2*2): #pool master or not; current or not
        cnt_xen.append(0)

    xen_servers= get_db_entries("xen_servers")
    for server in xen_servers:
       idx=enum_xenver[server.version] + 2* server.current + 2*2* server.is_pool_master
       cnt_xen[idx]+=1
       if server.current:
          total_num_cpu_sockets[enum_hv['XenServer']]+=server.num_cpu_sockets

    xen_resString=''
    for i in range(len(enum_xenver)*2*2): #pool master or not; current or not
        theHex=hex(cnt_xen[i])[2:]
        xen_resString+=theHex+"|"
    xen_resString=xen_resString[:-1]

    ###### nodes join application_instances        For OS
    psql_command  ="/usr/bin/psql bpdb postgres -tq -c \"SELECT os,app_id,instance_id,n.node_no "
    psql_command +="FROM bp.nodes as n INNER JOIN bp.application_instances as a ON n.node_no=a.node_no "
    psql_command +="ORDER BY instance_id\" | grep -v ^$"       
    instances = commands.getoutput(psql_command).splitlines()

    enum_app,tmpCnt,keyholder ={},0,[]
    total_instances=0
    ## Need 2 iteration of loops.
    #  1st time: fill up enum_app
    #  2nd time: count them based on enum app
    for i in range(len(instances)):
        line=instances[i]
        words=line.split("|")
        os=int(words[0])
        app_id=int(words[1])

        first_hex=hex(os)[2:]
        if len(first_hex)==1: first_hex="0"+first_hex
        second_hex=hex(app_id)[2:]
        if len(second_hex)==1: second_hex="0"+second_hex

        key="%s%s"% (first_hex, second_hex)
        total_instances+=1
        if not key in keyholder:
            keyholder.append(key)
            enum_app[key]=tmpCnt
            tmpCnt+=1

    message +="%d current among %d found " %(n_current,len(vmware_esx_servers))
    for i in range(len(enum)):
        theKey =''
        for key,value in enum.items():
            if i==value: theKey = key
        message += "%s(%d,%d) " %( theKey, cnts[i], cnts_current[i])

    ####    
    cnts=[]        
    for i in range(len(enum_app)):
        cnts.append(0)

    for i in range(len(instances)):
        line=instances[i]
        words=line.split("|")
        os=int(words[0])
        app_id=int(words[1])
    
        first_hex=hex(os)[2:]
        if len(first_hex)==1: first_hex="0"+first_hex
        second_hex=hex(app_id)[2:]
        if len(second_hex)==1: second_hex="0"+second_hex

        key="%s%s"% (first_hex, second_hex)
        cnts[enum_app[key]] += 1

    
    ###### application_instances join application_lookup to figure out HyperV or NOT
    psql_command  ="/usr/bin/psql bpdb postgres -tq -c \"SELECT node_no,instance_id,ai.app_id,name "
    psql_command +="FROM bp.application_instances AS ai,bp.application_lookup AS al "
    psql_command +="WHERE ai.app_id=al.app_id and name NOT LIKE '%VMware%' ORDER BY node_no\" | grep -v ^$"
    appinstances =commands.getoutput(psql_command).splitlines()

    psql_command  ="/usr/bin/psql bpdb postgres -tq -c \"SELECT node_no,os,version,num_cpu_sockets "
    psql_command +="FROM bp.nodes "
    psql_command +="ORDER BY node_no\" | grep -v ^$"
    nodes =commands.getoutput(psql_command).splitlines()

    nAsset=0
    for node in nodes:
        values=node.split("|")
        node_no=int(values[0])
        num_cpu_socket=int(values[3])
        hasHVapp=False
        hasSystemMetadata=False
        for app in appinstances:
            appvalues=app.split("|")
            if node_no != int(appvalues[0]): continue
            if "System Metadata" in appvalues[3]:
                hasSystemMetadata=True
            elif "Hyper-V" in appvalues[3]:
                hasHVapp=True
                break
        if hasSystemMetadata: 
            nAsset+=1
            continue
        if hasHVapp:
            total_num_cpu_sockets[enum_hv['HyperV']]+=num_cpu_socket
        else:    
            total_num_cpu_sockets[enum_hv['Else']]+=num_cpu_socket
        
    CONFIG.nAsset=nAsset

    resString=''
    for i in range(len(enum_hv)):
        theKey=''
        for key,value in enum_hv.items():
            if value==i: theKey=key
        theHex=hex(total_num_cpu_sockets[enum_hv[theKey]])[2:]
        #print theKey, total_num_cpu_sockets[enum_hv[theKey]]

        #nHexDigit=3
        #if theKey=='Else': nHexDigit=5
        #for i in range(nHexDigit-len(theHex)):
        #    theHex="0"+theHex
        resString+=theHex+"|"
    resString=resString[:-1]
    
    # Moved due to the need to get nAsset        
    cnts[enum_app['1401']]-=nAsset
    #message += " [INFO] Application_Version_Census "    
    message += " [INFO] AppVer "    
    for i in range(len(enum_app)):
        theKey =''
        for key,value in enum_app.items():
            if i==value: theKey = key

        if theKey=="141e" or theKey=="143c": continue # archive and SystemMetaData
        if cnts[enum_app[theKey]]==0: continue
        
        cnt_hex=hex(cnts[enum_app[theKey]])[2:]
        if CONFIG.debug: print "%s %4d %s" %(theKey,cnts[enum_app[theKey]],cnt_hex)
        if i!=0: message+="|"
        message += "%s%s"%(theKey,cnt_hex)
        
    message+="*"

    message += " [INFO] Sockets %s*" % resString    

    message += " [INFO] XenS %s" % xen_resString

#        if CONFIG.debug:
#            print "%2d, %3d, %s, %s"% (os,count,format(os,'02x'),format(count,'02x'))
#
#        if i!=0: message+="|"
#        message+=format(os,'02x')+format(count,'02x')
        

    return message

def getAnaProtected():
    message = "[INFO] Status of protection "
    if not CONFIG.db_accessible:
       message += " DB is not accessible because %s" % CONFIG.dbOff_reason 
       return message
    
    application_instances = get_db_entries("application_instances")
    backup_reps = get_db_entries("backup_reps") # collecting backups of which properties > 0
    for bckup in backup_reps:
        bckup.fillAttribute(application_instances) # pointer
        
    protected_instances = get_protected_instances(backup_reps)
    
    now = int(time.time())
    sums=[]
    cnts_found=[]
    cnts_currnt=[]
    cnts_proted=[]
    cnts_synced=[]
    import commands      
    psql_command = "/usr/bin/psql bpdb postgres -tq -c \"SELECT type FROM (SELECT * FROM bp.application_lookup ORDER BY app_id) AS a GROUP BY type\" | grep -v ^$"       
    if CONFIG.debug:
        print "#%s:getAnaProtected()" % basename
        print "     ",  psql_command
    app_types = commands.getoutput(psql_command).splitlines()

    enum= {}
    tmpCnt =0
    for i  in range(len(app_types)):
        app_type = app_types[i].strip()
        if app_type == "System Metadata" or app_type=="Archive": continue
        enum[app_type] = tmpCnt
        tmpCnt +=1

    for i in range(len(enum)):
        sums.append(0)
        cnts_found.append(0)
        cnts_currnt.append(0)
        cnts_proted.append(0)
        cnts_synced.append(0)


    application_lookup = get_db_entries("application_lookup")
    for app in application_instances:
        app_id = app.app_id
        app_type=get_application_type(application_lookup, app_id)
        if app_type == "System Metadata" or app_type=="Archive": continue
        cnts_found[enum[app_type]]+=1
        if app.sync:               cnts_synced[enum[app_type]]+=1
        if app.current:            cnts_currnt[enum[app_type]]+=1

    # loop over instances 
    for anInstance in protected_instances:
        app_id = anInstance[3]
        first_start_time = 0
        start_time = 0
        total_megs = 0
        node_type =-9
        os = ''
        type = ''
        # loop over backups including Master or Full which is the latest only
        for b in backup_reps:
            if not "Master" in b.backup_type and not "FULL" in b.backup_type: continue
            #if not b.sync: continue
            if anInstance[0]==b.node_no and anInstance[2]==b.instance_id:
                if start_time < b.start_time:
                    if start_time == 0:
                        first_start_time = b.start_time
                        node_type = b.node_type
                        os = b.os
                        type = b.backup_type
                    start_time = b.start_time
                    total_megs = b.total_megs
        app_type=get_application_type(application_lookup, app_id)
        if app_type == "System Metadata" or app_type=="Archive": continue
        cnts_proted[enum[app_type]]+=1
        sums[enum[app_type]]+=total_megs


    for key,value in enum.items():
        sums[value] = int(sums[value]/1024.+0.5)

    for i in range(len(enum)):
        theKey =''
        for key,value in enum.items():
            if i==value: theKey = key
        message += "[%s](%d,%d,%d,%d,%dG) " \
                   %(theKey,cnts_found[enum[theKey]],cnts_currnt[enum[theKey]],\
                     cnts_proted[enum[theKey]],cnts_synced[enum[theKey]],sums[enum[theKey]])


    # OS CENSUS    
    import commands      
    psql_command = "/usr/bin/psql bpdb postgres -tq -c \"SELECT os,count(os) FROM bp.nodes GROUP BY os ORDER BY os\" | grep -v ^$"       
    app_types = commands.getoutput(psql_command).splitlines()
    if CONFIG.debug:
        app_types=[]
        app_types.append(" 20 |    12")
        app_types.append(" 35 |     3")
        app_types.append(" 36 |    10")
        app_types.append(" 37 |    10")
        app_types.append(" 38 |    10")
        app_types.append(" 39 |    10")
        app_types.append(" 40 |    10")
        app_types.append(" 41 |     2")
        app_types.append(" 42 |     3")
            
    # How to convert hexadecimal to decimal: int('2a',16)
    message +=' [INFO] OS_Census '    
    for i in range(len(app_types)):
        line =app_types[i]
        words=line.split("|")
        os=int(words[0])
        count=int(words[1])
        if os==20: count-=CONFIG.nAsset   # deducting unitrends asset such as DPU or replication source
        if count==0: continue
        
        if i!=0: message+="|"
        #message+=format(os,'02x')+format(count,'02x')  # new in 2.6
        first_hex=hex(os)[2:]
        if len(first_hex)==1: first_hex="0"+first_hex
        second_hex=hex(count)[2:]
        #if len(second_hex)==1: second_hex="0"+second_hex

        if CONFIG.debug:
            print "%2d, %3d, %s, %s"% (os,count,first_hex,second_hex)

        message+=first_hex + second_hex

    message+="|ff%s*" %hex(CONFIG.nAsset)[2:]    


    # DedupNT
    bitmap =0
    if get_master_ini("fileDedup","MetadataGeneration")=="1":
        bitmap +=1
        if get_first_word("/usr/bp/bin/bputil -g fileDedup synchronizeUpdates ERROR /usr/bp/bpinit/master.ini").lower() == "no":
            bitmap +=2
    if get_first_word("/usr/bp/bin/bputil -g fileDedup enableInlineDedup ERROR /usr/bp/bpinit/master.ini").lower() == "yes":
        bitmap +=4
    
    nAgent_InlineDedup =0
    if int(get_first_word("/usr/bin/psql bpdb postgres -c \"select * from bp.application_instances limit 1\" | grep inline_dedup | wc -l")):
        nAgent_InlineDedup =\
		get_first_word("/usr/bin/psql bpdb postgres -tq -c \"select count(*) from bp.application_instances where inline_dedup!=0\" | grep -v ^$")
    if int(nAgent_InlineDedup): bitmap +=8
    message +=' [INFO] DedupNT %s' %hex(bitmap)[2:]


    return message


def get_application_type(application_lookup, app_id):
    for app in application_lookup:
        if app_id==app.app_id:
            return app.type
    return "NA"

#head -10 /usr/bp/logs.dir/vcd-worker-188.log | grep Hashing:137
def get_protected_instances(backup_reps):
    protected_instances=[]
    for b in backup_reps:
        if b.instance_id<=0 or b.app_id<0: continue
        aTuple = (b.node_no, b.node_name, b.instance_id, b.app_id, b.os, b.osid)
        found = False
        for node in protected_instances:
            if aTuple[0]==node[0] and aTuple[2]==node[2]:
                found = True
                break
        if not found: protected_instances.append( aTuple )
    return protected_instances    
    
def does_pass_threshold(shortName,days):
    """
    checks if elapsed days from the last time sending traps is longer than threshold (days)

    Keyword Arguments
    string shortName  : corresponds to 'object' in bp.snmp_history table to check the last time sending trap with the object
    int    days       : minimum days (threshold) to wait for sending the next one

    Returns:
    returns True if it is older than days
    """

    if not CONFIG.dbON:  # don't want to send a trap if db is down for analysis trap...
        return False
       #if CONFIG.dbON and os.path.exists("/var/log/monitoring_disks/bp.snmp_history.dbout"):
       #   command = "tac /var/log/monitoring_disks/bp.snmp_history.dbout | grep %s" % shortName 
    identifier =shortName
    psql_command ="/usr/bin/psql bpdb postgres -tq -c \""
    psql_command+="SELECT id,oid,time_sent,destination,community,traptype,regexp_replace(object,E'[\\n\\r]+',' ','g'),severity,status,description "
    psql_command+="FROM bp.snmp_history WHERE object LIKE '%%%s' ORDER BY id DESC LIMIT 1\" | grep -v ^$" % (identifier)
    import commands      
    import time, datetime

    if CONFIG.debug:
        print "#%s:does_pass_threshold()" % basename
        print "     ",  psql_command

    lines = commands.getoutput(psql_command).splitlines()
    if len(lines):
        words = lines[0].split("|")
        time_sent =get_time_stamp(words[2].strip())
        offset = days*24
        threshold = time_sent + datetime.timedelta(hours=offset)
        now = datetime.datetime.now()
        if now>threshold:
            return True
        else:
            return False
          
    return True  # if no entries found, then we have to send a trap

def get_time_stamp(stat_time):
    import datetime
    words = stat_time.split()
    #-rw-rw-r-- 1 root loggers     622 2013-07-30 13:29:12.324820223 -0400 /usr/bp/logs.dir/vcd-worker-5.log
    dateSegs = words[0].split("-")
    timeSegs = words[1].split(":")
    timestamp_start = datetime.datetime(int(dateSegs[0]), int(dateSegs[1]), int(dateSegs[2]), int(timeSegs[0]), int(timeSegs[1]), int(float(timeSegs[2])))
    return timestamp_start

if __name__ == '__main__':
    main()


